from ._version import __version__
from .actor import Actor
from .config import Configuration
from .proxy_configuration import ProxyConfiguration, ProxyInfo

__all__ = ['Actor', 'Configuration', 'ProxyConfiguration', 'ProxyInfo', '__version__']



from typing import TYPE_CHECKING, Dict, Optional, Protocol, Type, TypeVar, Union, cast

from apify_client import ApifyClientAsync

from ..config import Configuration
from ..memory_storage import MemoryStorage
from ..storage_client_manager import StorageClientManager

if TYPE_CHECKING:
    from .dataset import Dataset
    from .key_value_store import KeyValueStore
    from .request_queue import RequestQueue

T = TypeVar('T', 'Dataset', 'KeyValueStore', 'RequestQueue', covariant=True)


class Storage(Protocol[T]):
    """A protocol defining common interface for storage classes."""

    @classmethod
    def _create_instance(cls, storage_id_or_name: str, client: Union[ApifyClientAsync, MemoryStorage]) -> T:
        ...

    @classmethod
    def _get_default_name(cls, config: Configuration) -> str:
        ...


async def _purge_default_storages(client: Union[ApifyClientAsync, MemoryStorage]) -> None:
    if isinstance(client, MemoryStorage) and not client._purged:
        client._purged = True
        await client.purge()


class StorageManager:
    """A class for managing storages."""

    _default_instance: Optional['StorageManager'] = None
    _cache: Dict[Type[Storage], Dict[str, Storage]]

    def __init__(self) -> None:
        """Create a `StorageManager` instance."""
        self._cache = {}

    @classmethod
    def _get_default_instance(cls) -> 'StorageManager':
        if cls._default_instance is None:
            cls._default_instance = cls()

        return cls._default_instance

    @classmethod
    async def open_storage(
        cls,
        storage_class: Type[T],
        storage_id_or_name: Optional[str] = None,
        client: Optional[Union[ApifyClientAsync, MemoryStorage]] = None,
        config: Optional[Configuration] = None,
    ) -> T:
        """Open a storage of the given class, or return a cached storage object if it was opened before.

        Opens a new storage (`Dataset`, `KeyValueStore`, or `RequestQueue`) with the given ID or name.
        Returns the cached storage object if the storage was opened before.

        Args:
            storage_class (Type[Dataset] or Type[KeyValueStore] or Type[RequestQueue]): Class of the storage to be opened.
            storage_id_or_name (str, optional): ID or name of the storage to be opened. If omitted, an unnamed storage will be opened.
            client (ApifyClientAsync or MemoryStorage, optional): The storage client which should be used in the storage.
                If omitted, the default client will be used.
            config (Configuration, optional): The actor configuration to be used in this call. If omitted, the global configuration will be used.

        Returns:
            An instance of the storage given by `storage_class`.
        """
        storage_manager = StorageManager._get_default_instance()
        used_config = config or Configuration.get_global_configuration()
        used_client = client or StorageClientManager.get_storage_client()

        # Create cache for the given storage class if missing
        if storage_class not in storage_manager._cache:
            storage_manager._cache[storage_class] = {}

        # Fetch default name
        if not storage_id_or_name:
            storage_id_or_name = storage_class._get_default_name(used_config)

        # Try to get the storage instance from cache
        storage = storage_manager._cache[storage_class].get(storage_id_or_name, None)
        if storage is not None:
            # This cast is needed since we're storing all storages in one union dictionary
            return cast(T, storage)

        # Purge default storages if configured
        if used_config.purge_on_start:
            await _purge_default_storages(used_client)

        # Create the storage
        storage = await storage_class._create_instance(storage_id_or_name, used_client)

        # Cache by id and name
        storage_manager._cache[storage_class][storage._id] = storage
        if storage._name is not None:
            storage_manager._cache[storage_class][storage._name] = storage
        return storage

    @classmethod
    async def close_storage(cls, storage_class: Type[Storage], id: str, name: Optional[str]) -> None:
        """Close the given storage by removing it from the cache.

        Args:
            storage_class (Type[Dataset] or Type[KeyValueStore] or Type[RequestQueue]): Class of the storage to be closed.
            id (str): ID of the storage to be closed.
            name (str, optional): Name of the storage to be closed.
        """
        storage_manager = StorageManager._get_default_instance()
        del storage_manager._cache[storage_class][id]
        if name is not None:
            del storage_manager._cache[storage_class][name]



from operator import itemgetter
from typing import TYPE_CHECKING, Dict, Optional

from apify_client._utils import ListPage

from ..file_storage_utils import _update_metadata
from .key_value_store import KeyValueStoreClient, _find_or_cache_key_value_store_by_possible_id

if TYPE_CHECKING:
    from ..memory_storage import MemoryStorage


class KeyValueStoreCollectionClient:
    """Sub-client for manipulating key-value stores."""

    _key_value_stores_directory: str
    _client: 'MemoryStorage'

    def __init__(self, *, base_storage_directory: str, client: 'MemoryStorage') -> None:
        """Initialize the KeyValueStoreCollectionClient with the passed arguments."""
        self._key_value_stores_directory = base_storage_directory
        self._client = client

    def list(self) -> ListPage:
        """List the available key-value stores.

        Returns:
            ListPage: The list of available key-value stores matching the specified filters.
        """
        def map_store(store: KeyValueStoreClient) -> Dict:
            return store.to_key_value_store_info()
        return ListPage({
            'total': len(self._client._key_value_stores_handled),
            'count': len(self._client._key_value_stores_handled),
            'offset': 0,
            'limit': len(self._client._key_value_stores_handled),
            'desc': False,
            'items': sorted(map(map_store, self._client._key_value_stores_handled), key=itemgetter('createdAt')),
        })

    async def get_or_create(self, *, name: Optional[str] = None, _schema: Optional[Dict] = None) -> Dict:
        """Retrieve a named key-value store, or create a new one when it doesn't exist.

        Args:
            name (str, optional): The name of the key-value store to retrieve or create.
            schema (Dict, optional): The schema of the key-value store

        Returns:
            dict: The retrieved or newly-created key-value store.
        """
        if name:
            found = _find_or_cache_key_value_store_by_possible_id(client=self._client, entry_name_or_id=name)

            if found:
                return found.to_key_value_store_info()

        new_store = KeyValueStoreClient(name=name, base_storage_directory=self._key_value_stores_directory, client=self._client)
        self._client._key_value_stores_handled.append(new_store)

        kv_store_info = new_store.to_key_value_store_info()

        # Write to the disk
        await _update_metadata(data=kv_store_info, entity_directory=new_store._key_value_store_directory, write_metadata=self._client._write_metadata)

        return kv_store_info
